#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <assert.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <libssh/libssh.h>
#include <libssh/server.h>
#include "worker.h"
#include "globals.h"
#include "cmds.h"

char          printbuf[BUFSIZE];

char         *
wputs( char *dest, const char *str )
{
  int           l = strlen( str );
  if( l )
  {
    strncpy( dest, str, l );
    dest += l;
  }
  *dest++ = '\r';
  *dest++ = '\n';
  return dest;
}

char         *
wprintf( char *dest, char *fmt, ... )
{
  va_list       ap;
  int           l;

  va_start( ap, fmt );
  l = vsnprintf( dest, BUFSIZE, fmt, ap );
  va_end( ap );
  dest += l;
  *dest++ = '\r';
  *dest++ = '\n';
  return dest;
}

char         *
fixed_reply( char *dest, const char *lines[] )
{
  int           i, l;
  for( i = 0; lines[i]; i++ )
  {
    l = strlen( lines[i] );
    if( l )
    {
      strncpy( dest, lines[i], l );
      dest += l;
    }
    *dest++ = '\r';
    *dest++ = '\n';
  }
  return dest;
}

char         *
flush_delay( ssh_channel chan, char *d, int sec )
{
  ssh_channel_write( chan, printbuf, d - printbuf );
  sleep( sec );
  return printbuf;
}

#define TOTAL_FLG 0x80
#define HUMAN_FLG 0x40
#define MEGA_FLG 0x20
#define GIGA_FLG 0x10
#define UNAME_S "Linux "
#define UNAME_N "localhost "
#define UNAME_R "5.10.0-generic "
#define UNAME_V "#7 SMP PREEMPT SMP Debian 5.10.179-1 (2023-05-12) "
#define UNAME_M "x86_64 "
#define UNAME_O "GNU/Linux "

char          free_nr[MAXPATH];

char         *
fmt_nr( int free_flg, int val )
{
  static char  *p = free_nr;
  char         *here, *unit;

  unit = "";
  if( free_flg & HUMAN_FLG )
  {
    if( ( val + 500000 ) > 1000000 )
    {
      unit = "Gi";
      val = ( val + 500000 ) / 1000000;
    }
    else if( ( val + 500 ) > 1000 )
    {
      unit = "Mi";
      val = ( val + 500 ) / 1000;
    }
    else
      unit = "B";
  }
  else if( free_flg & MEGA_FLG )
    val = ( val + 500 ) / 1000;
  else if( free_flg & GIGA_FLG )
    val = ( val + 500000 ) / 1000000;
  if( p > free_nr + MAXPATH - 32 )
    p = free_nr;
  here = p;
  p += sprintf( here, "%d%s", val, unit );
  p++;
  return here;
}

int
parse_cmd( struct connection_info_t *conn, ssh_channel chan )
{
  int           i, l, cpu, free_flg;
  int           terminate = 0;
  int           maxcmds = 50;
  char         *p, *d, *cmd, *ccmd, *ecmd;
  float         bogo;

  ccmd = conn->cmdline;
  p = ccmd;
  while( *p )
  {
    if( *p == '&' )
      *p = ';';
    p++;
  }
  do
  {
    char         *str_st;
    int           str_state = ' ';
    if( ( ecmd = strchr( ccmd, ';' ) ) != NULL )
      *ecmd = '\0';
    memset( conn->cmdargs, 0, ( MAXARGS + 1 ) * sizeof( char * ) );
    p = ccmd;
    l = 0;
    // split the command in arguments
    while( *p && l < MAXARGS )
    {
      switch ( str_state )
      {
        case ' ':              // consuming spaces
          if( *p == '\"' )
          {
            str_st = p + 1;
            str_state = '\"';   // begin quote
          }
          else if( *p == '\'' )
          {
            str_st = p + 1;
            str_state = '\'';   // begin single quote
          }
          else if( *p != ' ' && *p != '\t' && *p != '\r' )
          {
            str_st = p;
            str_state = 'T';
          }
          break;
        case 'T':              // non-quoted text
          if( *p == ' ' || *p == '\t' || *p == '\r' || *p == '\"'
              || *p == '\'' )
          {
            *p = 0;
            conn->cmdargs[l++] = str_st;
            str_state = ' ';
          }
          else if( *p == '\"' )
            str_state = '\"';   // begin quote
          else if( *p == '\'' )
            str_state = '\'';   // begin single quote
          break;
        case '\"':             // inside a quote
          if( *p == '\"' )
          {
            str_state = 'T';    // end quote
            p--;
          }
          break;
        case '\'':             // inside a single quote
          if( *p == '\'' )
          {
            str_state = 'T';    // end single quote
            p--;
          }
          break;
      }
      p++;
    }                           // end while
    if( str_state != ' ' )
    {
      *p = 0;
      conn->cmdargs[l++] = str_st;
    }

    if( !--maxcmds )
      terminate = 1;
    if( conn->cmdargs[0] == NULL )
    {
      if( ecmd )
        ccmd = ecmd + 1;
      continue;                 // return 0;
    }
    if( ( cmd = strrchr( conn->cmdargs[0], '/' ) ) == NULL )
      cmd = conn->cmdargs[0];
    else
      cmd++;
    // handle commands
    d = printbuf;
    if( conn->cmdargs[1] && strstr( conn->cmdargs[1], "cpuinfo" ) )
    {
      for( cpu = 0; cpu < 8; cpu++ )
      {
        for( i = 0; cpuinfo[i][0]; i++ )
        {
          switch ( i )
          {
            case 0:
              d = wprintf( d, "%-16s: %d", cpuinfo[i][0], cpu );
              break;
            case 8:
              d = wprintf( d, "%-16s: %d", cpuinfo[i][0], ( cpu + 1 ) & 1 );
              break;
            case 10:
              d = wprintf( d, "%-16s: %d", cpuinfo[i][0], cpu >> 1 );
              break;
            case 23:
              bogo = atoi( cpuinfo[i][1] );
              bogo += ( ( random(  ) & 0xff ) - 0x80 ) / 100.;
              d = wprintf( d, "%-16s: %.2f", cpuinfo[i][0], bogo );
              break;
            default:
              d = wprintf( d, "%-16s: %s", cpuinfo[i][0], cpuinfo[i][1] );
              break;
          }
        }
        *d++ = '\r';
        *d++ = '\n';
      }
    }
    else if( !strcmp( cmd, "lscpu" ) )
    {
      for( i = 0; lscpu[i][0]; i++ )
      {
        d = wprintf( d, "%-21s%s", lscpu[i][0], lscpu[i][1] );
      }
    }
    else if( !strcmp( cmd, "free" ) )
    {
      free_flg = 0;
      if( conn->cmdargs[1] != NULL )
      {
        if( strchr( conn->cmdargs[1], 't' ) != NULL )
          free_flg |= TOTAL_FLG;
        if( strchr( conn->cmdargs[1], 'h' ) != NULL )
          free_flg |= HUMAN_FLG;
        if( strchr( conn->cmdargs[1], 'm' ) != NULL )
          free_flg |= MEGA_FLG;
        if( strchr( conn->cmdargs[1], 'g' ) != NULL )
          free_flg |= GIGA_FLG;
      }
      d = wprintf( d, "%7s%12s%12s%12s%12s%12s%12s",
          "", "total", "used", "free", "shared", "buff/cache", "available" );
      d = wprintf( d, "%-7s%12s%12s%12s%12s%12s%12s", "Mem:",
          fmt_nr( free_flg, 62644956 ),
          fmt_nr( free_flg, 15961588 ),
          fmt_nr( free_flg, 40163588 ),
          fmt_nr( free_flg, 113184 ),
          fmt_nr( free_flg, 4119780 ), fmt_nr( free_flg, 40347424 ) );
      d = wprintf( d, "%-7s%12s%12s%12s", "Swap:",
          fmt_nr( free_flg, 8388604 ),
          fmt_nr( free_flg, 0 ), fmt_nr( free_flg, 8388604 ) );
      if( free_flg & TOTAL_FLG )
        d = wprintf( d, "%-7s%12s%12s%12s", "Total:",
            fmt_nr( free_flg, 71033560 ),
            fmt_nr( free_flg, 15961588 ), fmt_nr( free_flg, 49515780 ) );
    }
    else if( !strcmp( cmd, "pwd" ) )
    {
      d = wputs( d, conn->pwd );
    }
    else if( !strcmp( cmd, "cd" ) )
    {
      if( conn->cmdargs[1] != NULL )
      {
        if( !strcmp( conn->cmdargs[1], ".." ) )
        {
          if( ( p = strrchr( conn->pwd, '/' ) ) != NULL )
          {
            if( p == conn->pwd )
              p++;
            *p = '\0';
          }
        }
        else
        {
          if( *conn->cmdargs[1] == '/' )
          {
            strncpy( conn->pwd, conn->cmdargs[1], MAXPATH - 1 );
            conn->pwd[MAXPATH - 1] = '\0';
          }
          else
          {
            p = conn->pwd + strlen( conn->pwd ) - 1;
            if( *p != '/' )
              strncat( conn->pwd, "/", MAXPATH );
            strncat( conn->pwd, conn->cmdargs[1], MAXPATH - 1 );
          }
        }
      }
      else
        strcpy( conn->pwd, "/root" );
    }
    else if( !strcmp( cmd, "busybox" ) )
    {
      d = fixed_reply( d, busybox_hdr );
      if( conn->cmdargs[1] == NULL )
        d = fixed_reply( d, busybox_usage );
    }
    else if( !strcmp( cmd, "gcc" ) )
    {
      if( conn->cmdargs[1] && !strcmp( conn->cmdargs[1], "--version" ) )
      {
        d = wputs( d, "gcc (Debian 8.3.0-2) 8.3.0" );
        d = wputs( d, "Copyright (C) 2018 Free Software Foundation, Inc." );
      }
      else
      {
        d = wputs( d, "gcc: fatal error: no input files" );
        d = wputs( d, "compilation terminated." );
      }
    }
    else if( !strcmp( cmd, "netstat" ) )
    {
      d = fixed_reply( d, netstat );
      d = wprintf( d,
          "tcp	 0     24 %s:%d	       %s:%d	 ESTABLISHED",
          conn->sockip, conn->dport, conn->peerip, conn->sport );
    }
    else if( !strcmp( cmd, "wget" ) )
    {
      d = fixed_reply( d, wget );
    }
    else if( !strcmp( cmd, "df" ) )
    {
      d = fixed_reply( d, df );
    }
    else if( !strcmp( cmd, "echo" ) )
    {
      unsigned char c;
      int           first_arg, newline, escapes;
      newline = 1;
      escapes = 0;
      for( i = 1; p = conn->cmdargs[i]; i++ )
      {
        if( conn->cmdargs[i][0] != '-' )
          break;
        if( conn->cmdargs[i][1] == 'e' )
          escapes = 1;
        if( conn->cmdargs[i][1] == 'n' )
          newline = 0;
      }
      first_arg = i;
      for( ; p = conn->cmdargs[i]; i++ )
      {
        if( i != first_arg )
          *d++ = ' ';
        if( escapes && strchr( p, '\\' ) )
        {
          while( ( c = *p++ ) )
          {
            if( c == '\\' && *p )
            {
              switch ( c = *p++ )
              {
                case 'a':
                  c = '\a';
                  break;
                case 'b':
                  c = '\b';
                  break;
                case 'e':
                  c = '\x1b';
                  break;
                case 'f':
                  c = '\f';
                  break;
                case 'n':
                  c = '\n';
                  break;
                case 'r':
                  c = '\r';
                  break;
                case 't':
                  c = '\t';
                  break;
                case 'v':
                  c = '\v';
                  break;
                case 'x':
                  {
                    unsigned char cl, ch = *p;
                    if( !isxdigit( ch ) )
                      goto not_an_escape;
                    p++;
                    c = ( ch | 0x20 ) - '0';
                    if( c > 9 )
                      c -= 0x27;
                    if( isxdigit( *p ) )
                    {
                      cl = ( *p | 0x20 ) - '0';
                      if( cl > 9 )
                        cl -= 0x27;
                      c = ( c << 4 ) | cl;
                      p++;
                    }
                  }
                  break;
                case '0':
                  c = 0;
                  if( !( '0' <= *p && *p <= '7' ) )
                    break;
                  c = *p++;
                  // FALLTHROUGH
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                  c -= '0';
                  if( '0' <= *p && *p <= '7' )
                    c = c * 8 + ( *p++ - '0' );
                  if( '0' <= *p && *p <= '7' )
                    c = c * 8 + ( *p++ - '0' );
                  break;
                case '\\':
                  break;
                not_an_escape:
                default:
                  c = '\\';
                  break;
              }
            }
            if( d < printbuf + BUFSIZE - 3 )
              *d++ = c;
          }                     // while
        }
        else
        {
          int           n = BUFSIZE - ( d - printbuf ) - 3;
          d = stpncpy( d, p, n );
        }
      }
      if( newline )
      {
        *d++ = '\r';
        *d++ = '\n';
      }
    }
    else if( !strcmp( cmd, "ifconfig" ) )
    {
      d = wputs( d, "eth0\tLink encap:Ethernet	HWaddr 00:30:66:29:50:83" );
      d = wprintf( d, "\tinet addr:%s  Bcast:%s	 Mask:255.255.255.0",
          conn->sockip, conn->sockip );
      d = fixed_reply( d, ifconfig_end );
    }
    else if( !strcmp( cmd, "ps" ) )
    {
      if( conn->cmdargs[1] == NULL )
      {
        d = fixed_reply( d, ps );
      }
      else
      {
        d = fixed_reply( d, ps_a );
      }
    }
    else if( !strcmp( cmd, "w" ) )
    {
      d = wputs( d,
          " 19:54:24 up 122 days,  6:02,  2 users,  load average: 0.08, 0.04, 0.05" );
      d = wputs( d,
          "USER	    TTY	     FROM	      LOGIN@   IDLE   JCPU   PCPU WHAT" );
      d = wputs( d,
          "root	    tty1		      Tue01   39:32m  2.12s  0.01s -bash" );
      d = wprintf( d, "root	pts/3	%-17s 19:54    1.00s  0.00s  0.00s w",
          conn->peerip );
    }
    else if( !strcmp( cmd, "whoami" ) )
    {
      d = wputs( d, "root" );
    }
    else if( !strcmp( cmd, "ls" ) || !strcmp( cmd, "ll" ) )
    {
      if( !strcmp( cmd, "ls" ) && ( conn->cmdargs[1] == NULL
          || *conn->cmdargs[1] != '-' ) )
        d = fixed_reply( d, ls );
      else
        d = fixed_reply( d, ls_l );
    }
    else if( !strcmp( cmd, "id" ) )
    {
      d = wputs( d,
          "uid=0(root) gid=0(root) groups=0(root),1(bin),2(daemon),3(sys),4(adm),6(disk),10(wheel)" );
    }
    else if( !strncmp( cmd, "apt", 3 ) )
    {
      if( conn->cmdargs[1] == NULL )
      {
        d = fixed_reply( d, apt );
      }
      else if( !strcmp( conn->cmdargs[1], "install" ) )
      {
        int           pkgsize = ( int ) random(  ) % 1500 + 101;
        char          pkgver[16];

        sprintf( pkgver, "%d.%d.%d-%d", ( int ) random(  ) % 4 + 1,
            ( int ) random(  ) % 9 + 1, ( int ) random(  ) % 9 + 1,
            ( int ) random(  ) % 4 + 1 );
        d = wputs( d, "Reading package lists... Done" );
        d = wputs( d, "Building dependency tree" );
        d = wputs( d, "Reading state information... Done" );
        d = wputs( d, "The following NEW packages will be installed:" );
        d = wprintf( d, "  %s", conn->cmdargs[2] );
        d = wputs( d,
            "0 upgraded, 1 newly installed, 0 to remove and 0 not upgraded." );
        d = wprintf( d, "Need to get %d kB of archives.", pkgsize );
        d = wprintf( d,
            "After this operation, %d kB of additional disk space will be used.",
            pkgsize * 2 );
        d = wprintf( d,
            "Get:1 http://ftp.us.debian.org/debian/ stretch/main %s amd64 %s [%d kB]",
            conn->cmdargs[2], pkgver, pkgsize );
        d = flush_delay( chan, d, 1 );
        d = wprintf( d, "Fetched %d kB in 0s (915 kB/s)", pkgsize );
        d = wprintf( d, "Selecting previously unselected package %s.",
            conn->cmdargs[2] );
        d = wputs( d,
            "(Reading database ... 24572 files and directories currently installed.)" );
        d = flush_delay( chan, d, 1 );
        d = wprintf( d, "Unpacking %s (from .../%s_%s_amd64.deb) ...",
            conn->cmdargs[2], conn->cmdargs[2], pkgver );
        d = flush_delay( chan, d, 2 );
        d = wputs( d, "Processing triggers for man-db ..." );
        d = flush_delay( chan, d, 3 );
        d = wprintf( d, "Setting up %s (%s) ...", conn->cmdargs[2], pkgver );
      }
      else
        d = wputs( d, "E: Invalid operation" );
    }
    else if( !strcmp( cmd, "uname" ) )
    {
      union
      {
        unsigned      all;
        struct
        {
          unsigned      s:1;
          unsigned      n:1;
          unsigned      r:1;
          unsigned      v:1;
          unsigned      m:1;
          unsigned      o:1;
        } sw;
      } uname;
      uname.all = 0;
      if( conn->cmdargs[1] == NULL )
        uname.sw.s = 1;
      else
      {
        for( l = 1; l < MAXARGS; l++ )  // log the command
        {
          if( conn->cmdargs[l] == NULL )
            break;
          if( conn->cmdargs[l][0] != '-' )
            continue;
          switch ( conn->cmdargs[l][1] )
          {
            case 'a':
              uname.all = -1;
              break;
            case 's':
              uname.sw.s = 1;
              break;
            case 'n':
              uname.sw.n = 1;
              break;
            case 'r':
              uname.sw.r = 1;
              break;
            case 'v':
              uname.sw.v = 1;
              break;
            case 'm':
              uname.sw.m = 1;
              break;
            case 'o':
              uname.sw.o = 1;
              break;
          }
        }
      }
      printbuf[0] = 0;
      if( uname.sw.s )
        strcat( printbuf, UNAME_S );
      if( uname.sw.n )
        strcat( printbuf, UNAME_N );
      if( uname.sw.r )
        strcat( printbuf, UNAME_R );
      if( uname.sw.v )
        strcat( printbuf, UNAME_V );
      if( uname.sw.m )
        strcat( printbuf, UNAME_M );
      if( uname.sw.o )
        strcat( printbuf, UNAME_O );
      d = printbuf + strlen( printbuf ) - 1;
      *d++ = '\r';
      *d++ = '\n';
    }
    else if( !strcmp( cmd, "exit" ) || !strcmp( cmd, "logout" ) )
    {
      d = wputs( d,
          ">>> Your pathetic hacking attempt session has been logged <<<" );
      terminate = 1;
    }
    else if( !strcmp( cmd, "make" ) )
    {
      d = wputs( d,
          "make: *** No targets specified and no makefile found.  Stop." );
    }
    else if( !strcmp( cmd, "nproc" ) )
    {
      d = wputs( d, "16" );
    }
    else if( !strcmp( cmd, "uptime" ) )
    {
      d = wputs( d,
          "17:08:55 up 139 days,  3:13,	 6 users,  load average: 4.12, 5.29, 3.78" );
    }
    else if( !strcmp( cmd, "cp" ) || !strcmp( cmd, "mv" )
        || !strcmp( cmd, "rm" ) || !strcmp( cmd, "cat" )
        || !strcmp( cmd, "more" ) || !strcmp( cmd, "grep" )
        || !strcmp( cmd, "chown" ) || !strcmp( cmd, "chmod" )
        || !strcmp( cmd, "kill" ) || !strcmp( cmd, "killall" ) )
    {
      // do nothing
    }
    else
    {
      d = wprintf( d, "-bash: %s: command not found", conn->cmdargs[0] );
    }
    ssh_channel_write( chan, printbuf, d - printbuf );
    p = conn->buf;
    for( l = 0; l < MAXARGS; l++ )      // log the command
    {
      if( conn->cmdargs[l] == NULL )
        break;
      p += sprintf( p, "%s ", conn->cmdargs[l] );
    }
    if( p > conn->buf )
      *( p - 1 ) = '\n';
    fprintf( globals.flog, "%s", conn->buf );
    fflush( globals.flog );
    if( ecmd )
      ccmd = ecmd + 1;
  }
  while( ecmd && !terminate );
  return terminate;
}

char         *
my_time( const struct tm *timeptr )
{
  static char   mon_name[12][3] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
  };
  static char   result[32];

  sprintf( result, "%d-%.3s-%d %.2d:%.2d:%.2d",
      timeptr->tm_mday,
      mon_name[timeptr->tm_mon],
      1900 + timeptr->tm_year,
      timeptr->tm_hour, timeptr->tm_min, timeptr->tm_sec );
  return result;
}

void         *
worker( void *arg )
{
  struct sockaddr_storage addr;
  ssh_message   message = NULL;
  ssh_channel   chan = 0;
  int           l;
  int           auth = 0;
  int           shell = 0;
  int           terminate = 0;
  char         *u, *p, *ppwd, *cmd;
  struct tm    *t_ptr;
  time_t        lt;

  struct connection_info_t *conn = ( struct connection_info_t * ) arg;

  ssh_session   session = conn->session;
  socket_t      sock = ssh_get_fd( session );
  int           version = ssh_get_version( session );
  socklen_t     len = sizeof( addr );

  if( !getsockname( sock, ( struct sockaddr * ) &addr, &len ) )
  {
    if( addr.ss_family == AF_INET )
    {
      struct sockaddr_in *s = ( struct sockaddr_in * ) &addr;

      conn->dport = ntohs( s->sin_port );
      inet_ntop( AF_INET, &s->sin_addr, conn->sockip,
          sizeof( conn->sockip ) );
    }
    else if( addr.ss_family == AF_INET6 )
    {                           // AF_INET6
      struct sockaddr_in6 *s = ( struct sockaddr_in6 * ) &addr;

      conn->dport = ntohs( s->sin6_port );
      inet_ntop( AF_INET6, &s->sin6_addr, conn->sockip,
          sizeof( conn->sockip ) );
    }
    else
    {
      /* 
       * Should not happen
       */
      assert( 0 );
    }
  }
  else
  {
    conn->sockip[0] = '?';
    conn->sockip[1] = 0;
    conn->dport = -1;
  }
  if( !getpeername( sock, ( struct sockaddr * ) &addr, &len ) )
  {
    if( addr.ss_family == AF_INET )
    {
      struct sockaddr_in *s = ( struct sockaddr_in * ) &addr;

      conn->sport = ntohs( s->sin_port );
      inet_ntop( AF_INET, &s->sin_addr, conn->peerip,
          sizeof( conn->peerip ) );
    }
    else if( addr.ss_family == AF_INET6 )
    {                           // AF_INET6
      struct sockaddr_in6 *s = ( struct sockaddr_in6 * ) &addr;

      conn->sport = ntohs( s->sin6_port );
      inet_ntop( AF_INET6, &s->sin6_addr, conn->peerip,
          sizeof( conn->peerip ) );
    }
    else
    {
      /* 
       * Should not happen
       */
      assert( 0 );
    }
  }
  else
  {
    conn->peerip[0] = '?';
    conn->peerip[1] = 0;
    conn->sport = -1;
  }

  if( SSH_OK == ssh_handle_key_exchange( session ) )
  {
#if LIBSSH_VERSION_INT >= SSH_VERSION_INT(0, 6, 0)
    ssh_set_auth_methods( session, SSH_AUTH_METHOD_PASSWORD );
#endif

    do
    {
      message = ssh_message_get( session );

      if( !message || globals.terminate )
      {
        break;
      }
      switch ( ssh_message_type( message ) )
      {
        case SSH_REQUEST_AUTH:
          switch ( ssh_message_subtype( message ) )
          {
            case SSH_AUTH_METHOD_PASSWORD:
              u = ( char * ) ssh_message_auth_user( message );
              p = ( char * ) ssh_message_auth_password( message );
              if( !strcmp( u, "root" ) && p && *p == '1' )
              {
                ssh_message_auth_reply_success( message, 0 );
                lt = time( NULL );
                t_ptr = localtime( &lt );
                fprintf( globals.flog, "From %s %s ssh%d %s/%s\n",
                    conn->peerip, my_time( t_ptr ), version, u, p );
                fflush( globals.flog );
                auth = 1;
              }
              else
              {
                ssh_message_reply_default( message );
#if 0
                fprintf( globals.flog, "Try %s port %d ssh%d %s/%s\n",
                    peerip, conn->sport, version, u, p );
                fflush( globals.flog );
#endif
              }
              break;
            default:
              ssh_message_auth_set_methods( message,
                  SSH_AUTH_METHOD_PASSWORD );
              ssh_message_reply_default( message );
              break;
          }
          break;

        default:
          ssh_message_reply_default( message );
          break;
      }
      ssh_message_free( message );
    }
    while( !globals.terminate && !auth );
  }
  if( auth )
  {
    do
    {
      message = ssh_message_get( session );
      if( message )
      {
        switch ( ssh_message_type( message ) )
        {
          case SSH_REQUEST_CHANNEL_OPEN:
            if( ssh_message_subtype( message ) == SSH_CHANNEL_SESSION )
            {
              chan = ssh_message_channel_request_open_reply_accept( message );
              break;
            }
          default:
            ssh_message_reply_default( message );
        }
        ssh_message_free( message );
      }
    }
    while( message && !chan );
    do
    {
      message = ssh_message_get( session );
      if( message && ssh_message_type( message ) == SSH_REQUEST_CHANNEL )
      {
        switch ( ssh_message_subtype( message ) )
        {
          case SSH_CHANNEL_REQUEST_SHELL:
            ssh_message_channel_request_reply_success( message );
            shell = 1;
            break;
          case SSH_CHANNEL_REQUEST_PTY:
            ssh_message_channel_request_reply_success( message );
            break;
          case SSH_CHANNEL_REQUEST_EXEC:
            ssh_message_channel_request_reply_success( message );
            cmd = ( char * ) ssh_message_channel_request_command( message );
            fprintf( globals.flog, "Exec %s\n", cmd );
            fflush( globals.flog );
            strncpy( conn->cmdline, cmd, BUFSIZE );
            conn->cmdline[BUFSIZE - 1] = '\0';
            parse_cmd( conn, chan );
            terminate = 1;
            break;
          default:
            ssh_message_reply_default( message );
            break;
        }
      }
      else
        ssh_message_reply_default( message );
      ssh_message_free( message );
    }
    while( message && !( shell || terminate ) );
    if( !terminate )
    {
      ssh_channel_write( chan, UNAME_S, strlen( UNAME_S ) );
      ssh_channel_write( chan, UNAME_N, strlen( UNAME_N ) );
      ssh_channel_write( chan, UNAME_R, strlen( UNAME_R ) );
      ssh_channel_write( chan, UNAME_V, strlen( UNAME_V ) );
      ssh_channel_write( chan, UNAME_M, strlen( UNAME_M ) );
      ssh_channel_write( chan, UNAME_O, strlen( UNAME_O ) );
      ssh_channel_write( chan, "\r\n", 2 );
    }
    strcpy( conn->pwd, "/root" );
    do
    {
      if( terminate )
        break;
      strcpy( conn->buf, "[root@" );
      p = conn->buf + strlen( conn->buf );
      gethostname( p, BUFSIZE - 7 );
      if( ( p = strchr( p, '.' ) ) != NULL )
        *p = '\0';
      if( ( ppwd = strrchr( conn->pwd, '/' ) ) == NULL )
        ppwd = conn->pwd;
      else if( *( ppwd + 1 ) && ppwd != conn->pwd )
        ppwd++;
      p = conn->buf + strlen( conn->buf );
      sprintf( p, " %s]# ", ppwd );
      ssh_channel_write( chan, conn->buf, strlen( conn->buf ) );
      cmd = conn->cmdline;
      while( 1 )
      {
        l = ssh_channel_read( chan, conn->buf, BUFSIZE, 0 );    // blocking
        if( l < 1 )
        {
          *cmd = '\0';
          terminate = 1;
          break;
        }
        // special character
        if( l >= 3 && conn->buf[0] == 0x1b && conn->buf[1] == 0x5b )
        {
          if( conn->buf[2] == 0x44 )    // arrow left
          {
            if( cmd > conn->cmdline )
            {
              ssh_channel_write( chan, conn->buf, l );
              cmd--;
            }
          }
          else if( conn->buf[2] == 0x43 )       // arrow right
          {
            if( *cmd && cmd < conn->cmdline + BUFSIZE - 1 )
            {
              ssh_channel_write( chan, conn->buf, l );
              cmd++;
            }
          }
          else if( conn->buf[2] == 0x41 )       // arrow up
          {
            p = conn->buf;
            for( l = 0; l < MAXARGS; l++ )
            {
              if( conn->cmdargs[l] == NULL )
                break;
              p += sprintf( p, "%s ", conn->cmdargs[l] );
            }
            l = p - conn->buf - 1;
            if( l > 0 )
            {
              strncpy( conn->cmdline, conn->buf, l );
              cmd = conn->cmdline + l;
              ssh_channel_write( chan, conn->buf, l );
            }
          }
          continue;
        }
        *cmd = '\0';
        if( conn->buf[0] == 0x7f )      // delete key
        {
          if( cmd > conn->cmdline )
          {
            conn->buf[0] = 8;
            conn->buf[1] = ' ';
            conn->buf[2] = 8;
            ssh_channel_write( chan, conn->buf, 3 );
            cmd--;
            *cmd = '\0';
          }
          continue;
        }
        ssh_channel_write( chan, conn->buf, l );
        if( conn->buf[0] == '\r' )
        {
          *cmd = '\0';
          break;
        }
        if( cmd + l < conn->cmdline + BUFSIZE )
        {
          strncpy( cmd, conn->buf, l );
          cmd += l;
        }
      }
      if( terminate )
        break;
      ssh_channel_write( chan, "\n", 1 );
      terminate = parse_cmd( conn, chan );
    }
    while( !terminate );
  }
  finalize_connection( conn );
  if( !globals.terminate )
  {
    pthread_detach( pthread_self(  ) );
  }
  return 0;
}

void
finalize_connection( struct connection_info_t *conn )
{
  ssh_session   session = conn->session;

  pthread_mutex_lock( &globals.mutex );
  {
    if( conn->prev )
    {
      conn->prev->next = conn->next;
    }

    if( conn->next )
    {
      conn->next->prev = conn->prev;
    }

    if( globals.tail == conn )
    {
      globals.tail = conn->prev;
    }

    if( globals.head == conn )
    {
      globals.head = conn->next;
    }

    --globals.n_threads;
  }
  pthread_mutex_unlock( &globals.mutex );

  ssh_disconnect( session );
  ssh_free( session );
  free( conn );
}
